package czerwony.kapturek.jbcolumn;

import java.util.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.datatransfer.*;
import javax.swing.*;
import javax.swing.text.*;
import com.borland.primetime.editor.*;
import com.borland.primetime.ide.*;

/**
 * @todo cursor is placed in strange places after block text operations
 * @todo sometimes undo works in a quite starange way
 * @todo action to insert a prefix before selected column
 */
class ColumnActions
{
//    private final static String TXT_ACTIONPREFIX = "jbBrief-";
    private final static String TXT_MYNAME          = "Czerwony Kapturek's Column Map";
    private final static String TXT_ACTIONPREFIX    = "jbColumn-";

    private final static String TXT_UP      = "up";
    private final static String TXT_DOWN    = "down";
    private final static String TXT_RIGHT   = "right";
    private final static String TXT_LEFT    = "left";
    private final static String TXT_ESCAPE  = "escape";

    private final static String TXT_COLOR = "Column block highlight";
    private final static Color  DEF_BACKGROUND = Color.green;

    private final static JTextComponent.KeyBinding columnMapFixedBindings[] = {
        new JTextComponent.KeyBinding(
                KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0), TXT_ACTIONPREFIX + TXT_ESCAPE),
        new JTextComponent.KeyBinding(
                KeyStroke.getKeyStroke(KeyEvent.VK_UP, 0), TXT_ACTIONPREFIX + TXT_UP),
        new JTextComponent.KeyBinding(
                KeyStroke.getKeyStroke(KeyEvent.VK_DOWN, 0), TXT_ACTIONPREFIX + TXT_DOWN),
        new JTextComponent.KeyBinding(
                KeyStroke.getKeyStroke(KeyEvent.VK_LEFT, 0), TXT_ACTIONPREFIX + TXT_LEFT),
        new JTextComponent.KeyBinding(
                KeyStroke.getKeyStroke(KeyEvent.VK_RIGHT, 0), TXT_ACTIONPREFIX + TXT_RIGHT),
    };


//    public EditorAction         action_BufferList = new BufferListAction();
    public EditorAction			action_StartColumnSelection = new StartColumnSelectionAction();
    public EditorAction			action_ColumnPaste = new ColumnPasteAction();
    public EditorAction         action_Test = new TestAction();

    public EditorAction			action_ColumnUp = new ColumnUpAction();
    public EditorAction			action_ColumnDown = new ColumnDownAction();
    public EditorAction			action_ColumnLeft = new ColumnLeftAction();
    public EditorAction			action_ColumnRight = new ColumnRightAction();
    public EditorAction			action_Escape = new EscapeAction();

    public EditorAction			action_ColumnCopy = new ColumnCopyAction();
    public EditorAction			action_ColumnCut = new ColumnCutAction();
    public EditorAction			action_ColumnDelete = new ColumnDeleteAction();
    public EditorAction			action_ColumnUpperCase = new ColumnUpperCaseAction();
    public EditorAction			action_ColumnLowerCase = new ColumnLowerCaseAction();
    public EditorAction         action_ColumnToggleCase = new ColumnToggleCaseAction();

    private int         startRow;
    private int         startColumn;
    private int         row;
    private int         column;
    private Object      highlight;
    private int         cx;
    private int         cy;
    private Rectangle   r = new Rectangle();
    private int         topMargin;
    private int         leftMargin;
    private Keymap      oldKeymap;
    private EditorPane  ed;
    private String      clipText;
    private boolean     clipOwner;

    public ColumnActions()
    {
        MasterStyleContext styleCtx = EditorManager.getStyleContext();
        Style style = styleCtx.addStyle(TXT_COLOR, null);
        style.addAttribute(Style.NameAttribute, TXT_COLOR);
        StyleConstants.setBackground(style, DEF_BACKGROUND);


        Keymap sub = EditorManager.createKeymap( TXT_MYNAME,
                                                 columnMapFixedBindings,
                                                 getColumnMapActions() );
        sub.setDefaultAction(action_Test);
        EditorManager.registerKeymap(sub);

    }

/////////////////////////////////////////////////////////////////////////////////////////////////
// actions defined by this map

    public EditorAction[] getActions()
    {
        EditorAction[] a = {action_StartColumnSelection,
                            action_ColumnPaste,
                            //action_BufferList
                           };
        return a;
    }


    private EditorAction[] getColumnMapActions()
    {
        EditorAction[] a = {action_Escape,
                            action_ColumnUp,
                            action_ColumnDown,
                            action_ColumnLeft,
                            action_ColumnRight};
        return a;
    }

//////////////////////////////////////////////////////////////////////////////////////////////////
// utility functions

    void moveSelection(int dx, int dy, EditorPane ep)
    {
        try
        {
            column += dx;
            row += dy;
            column = Math.max(0, column);
            row = Math.max(0, row);
            row = Math.min(row, ep.getLineCount()-1);

            Element line = ep.getDocument().getRootElements()[0];
            int firstLine = Math.min(startRow, row);
            int lastLine = Math.max(startRow, row);
            int s1 = line.getElement(firstLine).getStartOffset();
            int s2 = (lastLine+1 == ep.getLineCount())
                        ? ep.getDocument().getEndPosition().getOffset()
                        : line.getElement(lastLine+1).getStartOffset();

            if (highlight==null)
            {
                highlight = ep.getHighlighter().addHighlight(s1, s2, new BlockHighlightPainter(ep));
            }
            else
            {
                ep.getHighlighter().changeHighlight(highlight, s1, s2);
            }

            ep.scrollRectToVisible(new Rectangle(column*cx, row*cy, 1, 1));
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
        }
    }


    private void parseBlock(EditorPane ed, BlockParserListener listener)
    {
        int row1 = Math.min(startRow, row)+1;
        int row2 = Math.max(startRow, row)+1;
        int col1 = Math.min(startColumn, column);
        int col2 = Math.max(startColumn, column);
        Element root = ed.getDocument().getRootElements()[0];
        int required = col2-col1+1;
        int tabSize = EditorManager.getTabSize();
        StringBuffer sb = null;
        int offsetStart = 0;
        for (int i = row1; i <= row2; i++)
        {
            offsetStart = root.getElement(i-1).getStartOffset();
            char[] line = ed.getTextLine(i).toCharArray();
            int len = 0;
            int copied = 0;
            int skipped = 0;
            sb = new StringBuffer();
            for (int j = 0; (j < line.length) && (copied < required); j++)
            {
                char c = line[j];
                if ((c == '\r') || (c == '\n'))
                {
                    continue;
                }

                if (skipped < col1)
                {
                    if (c == '\t')
                    {
                        skipped += (tabSize - (skipped % tabSize));
                        for (int k=0; (k < (skipped-col1)) && (copied < required); k++)
                        {
                            copied++;
                            sb.append(' ');
                        }
                    }
                    else
                    {
                        skipped++;
                    }
                    offsetStart++;
                }
                else
                {
                    if (c == '\t')
                    {
                        int toFullTab = (tabSize - (copied % tabSize));
                        for (int k=0; (k < toFullTab) && (copied < required); k++)
                        {
                            copied++;
                            sb.append(' ');
                        }
                    }
                    else
                    {
                        copied++;
                        sb.append(c);
                    }
                    len++;
                }
            }

            try
            {
                String replace = listener.lineRead(sb.toString(), required, copied);
                if (replace != null)
                {
                    ed.getDocument().remove(offsetStart, len);
                    if (replace.length() > 0)
                    {
                        if (skipped < col1)
                        {
//                            System.err.println("skipped/col1" + skipped+"/"+col1);
                            StringBuffer tmp = new StringBuffer();
                            for (; skipped < col1; skipped++)
                            {
//                                System.err.println("adding space");
                                tmp.append(' ');
                            }
                            replace = tmp.toString() + replace;
                        }
                        ed.getDocument().insertString(offsetStart, replace, null);
//                        System.err.println("inserted: \"" + replace + "\"");
                    }
                }
            }
            catch (BadLocationException ex)
            {
                ex.printStackTrace();
            }
        }
    }


    private void cancelSelection()
    {
        ed.getCaret().setVisible(true);
        if (highlight != null)
        {
            ed.getHighlighter().removeHighlight(highlight);
        }
        ed.setKeymap(oldKeymap);
        resetData();
    }


    private void resetData()
    {
        startRow = 0;
        startColumn = 0;
        row = 0;
        column = 0;
        highlight = null;
        cx = 0;
        cy = 0;
        r = new Rectangle();
        topMargin = 0;
        leftMargin = 0;
        oldKeymap = null;
        ed = null;
        clipText = null;
        clipOwner = false;
    }


    /**
     *  Listener for parseBlock()
     */
    public interface BlockParserListener
    {
        public String lineRead(String line, int required, int actual);
    }


//////////////////////////////////////////////////////////////////////////////////////////////////
// action classes

    private abstract class BriefAction extends EditorAction
    {
        BriefAction(String name, String descr)
        {
            super(TXT_ACTIONPREFIX + name);
//            System.err.println("creating: " + TXT_ACTIONPREFIX + name);
            putValue(LONG_DESCRIPTION, descr);
            putValue("ActionGroup", "Column selection (jbColumn)" );
        }

        public void actionPerformed(ActionEvent ev)
        {
            EditorPane ed = getEditorTarget(ev);
            if(ed != null)
            {
                JBUndoManager undo = ((EditorDocument)ed.getDocument()).getUndoManager();
                Caret caret = ed.getCaret();
                undo.startUndoGroup(caret.getDot(), caret.getMark());
                onAction(ed);
                undo.endUndoGroup(caret.getDot(), caret.getMark());
            }
        }

        protected void onAction(EditorPane ed)
        {
        }
    }

    private DelayedAction delayedAction;

        //UGLY!!!: it looks like in Java there is no way to check what is the character coresponding
        // to given VK_ code. For instance there is no generic way to check what char (if any) will
        // VK_PLUS generate. KeyEvent instances for KEY_PRESSED events have valid char codes
        // but they are set directly from JNI. The funny thisng is that KeyEvent constructors
        // are actually not used - all parameters passed to them are zeros and null.
    private class DelayedAction extends EditorAction
    {
        Action        target;
        ActionEvent   ev;

        DelayedAction(Action target)
        {
            super("");
            this.target = target;
        }

        public void actionPerformed(ActionEvent ev)
        {
            this.ev = ev;
            delayedAction = this;
        }

        public void fire()
        {
            delayedAction = null;
            target.actionPerformed(ev);
        }
    }

//////////////////////////////////////////////////////////////////////////////////////////////////
// extermnal actions - to be mapped by the user

    private static int[] specialKeys =
        {KeyEvent.VK_ADD, KeyEvent.VK_SUBTRACT, KeyEvent.VK_MULTIPLY};

    public class StartColumnSelectionAction extends BriefAction
    {
        private Keymap newKeymap;


        public StartColumnSelectionAction()
        {
            super("start-selection", "");
            KeymapManager.installRegisteredKeymap(Browser.getActiveBrowser(), "Brief-Column-Map");
        }


        protected void onAction(EditorPane newEd)
        {
            if (ed != null)
                cancelSelection();
            ed = newEd;
            int dotPos = ed.getCaret().getDot();
            row = ed.getLineNumber(dotPos)-1;
            startRow = row;
            column = ed.getColumnNumber(dotPos)-1;
            startColumn = column;
            leftMargin = ed.getMargin().left;
            topMargin = ed.getMargin().top;
            cy = ed.getHeight() / ed.getLineCount();
            cx = ed.getFontWidth();
            oldKeymap = EditorManager.getKeymap();
            ed.getCaret().setVisible(false);
            newKeymap = EditorManager.getKeymap(TXT_MYNAME);

            remapAction(EditorActions.ACTION_Copy, action_ColumnCopy);
            remapAction(com.borland.primetime.editor.BriefActions.ACTION_BriefCopy, action_ColumnCopy);
            remapAction(EditorActions.ACTION_Cut, action_ColumnCut);
            remapAction(com.borland.primetime.editor.BriefActions.ACTION_BriefCut, action_ColumnCut);
            remapAction(EditorActions.ACTION_DeleteSelection, action_ColumnDelete);
            remapAction(com.borland.primetime.editor.BriefActions.ACTION_BriefDelete, action_ColumnDelete);
            remapAction(EditorActions.ACTION_UpperCase, action_ColumnUpperCase);
            remapAction(EditorActions.ACTION_LowerCase, action_ColumnLowerCase);
            remapAction(EditorActions.ACTION_ToggleCase, action_ColumnToggleCase);
            remapAction(action_StartColumnSelection, action_Escape);

            ed.setKeymap(newKeymap);
            moveSelection(0, 0, ed);
        }


        private void remapAction(Action oldAction, Action newAction)
        {
            KeyStroke[] k = oldKeymap.getKeyStrokesForAction(oldAction);

            if (k == null)
            {
                return;
            }

            for (int i = 0; i < k.length; i++)
            {
                if (isSpecialKey(k[i].getKeyCode()))
                {
                    newKeymap.addActionForKeyStroke(k[i],
                            new DelayedAction(newAction));
                }
                else
                {
                    newKeymap.addActionForKeyStroke(k[i], newAction);
                }
            }
        }


        private boolean isSpecialKey(int code)
        {
            for (int i = 0; i < specialKeys.length; i++)
            {
                if (specialKeys[i] == code)
                {
                    return true;
                }
            }
            return false;
        }
    }


    public class ColumnPasteAction extends BriefAction
            implements BlockParserListener
    {
        StringTokenizer st;


        public ColumnPasteAction()
        {
            super("paste", "");
        }


        public void actionPerformed(ActionEvent ev)
        {
            if (clipOwner)
            {
                super.actionPerformed(ev);
            }
            else
            {
                com.borland.primetime.editor.BriefActions.ACTION_BriefPaste.actionPerformed(ev);
            }
        }


        protected void onAction(EditorPane ed)
        {
            int lineLen = clipText.indexOf("\n");
            int clipLines = clipText.length() / (lineLen + 1);
            int dotPos = ed.getCaret().getDot();
            startRow = ed.getLineNumber(dotPos) - 1; //-1 because our coordinates are zero-based
            startColumn = ed.getColumnNumber(dotPos) - 1;
            row = startRow + clipLines - 1;
            column = EditorManager.isInsertMode() ? startColumn : startColumn + lineLen;
            st = new StringTokenizer(clipText, "\n");
            ensureEditorHasEnoughLines(ed, row + 1); //+1 because row is zero-based
            parseBlock(ed, this);
        }


        private void ensureEditorHasEnoughLines(EditorPane ed, int reqLines)
        {
            int edLines = ed.getLineCount();
            Document doc = ed.getDocument();
            for (int i = 0; i < (reqLines - edLines); i++)
            {
                try
                {
                    doc.insertString(doc.getLength(), "\n", null);
                }
                catch (BadLocationException ex)
                {
                    ex.printStackTrace();
                }
            }
        }


        public String lineRead(String line, int required, int actual)
        {
            String ret = st.nextToken();
            if ((line != null) && (line.length() > 0) && (required == actual))
            {
                ret += line.charAt(EditorManager.isInsertMode() ? 0 : actual-1);
            }
            return ret;
        }
    }


//////////////////////////////////////////////////////////////////////////////////////////////////
// Keymap's internal actions

    /**
     * Used as default action for column selection keymap.
     */
    public class TestAction extends BriefAction
    {
        public TestAction()
        {
            super("test", "");
        }

        public void actionPerformed(ActionEvent ev)
        {
            if (delayedAction != null)
                delayedAction.fire();
        }
    }


    public class ColumnUpAction extends BriefAction
    {
        public ColumnUpAction()
        {
            super(TXT_UP, "");
        }

        protected void onAction(EditorPane ed)
        {
            moveSelection(0, -1, ed);
        }
    }

    public class ColumnDownAction extends BriefAction
    {
        public ColumnDownAction()
        {
            super(TXT_DOWN, "");
        }

        protected void onAction(EditorPane ed)
        {
//            System.err.println("down");
            moveSelection(0, 1, ed);
        }
    }


    public class ColumnLeftAction extends BriefAction
    {
        public ColumnLeftAction()
        {
            super(TXT_LEFT, "");
        }

        protected void onAction(EditorPane ed)
        {
//            System.err.println("left");
            moveSelection(-1, 0, ed);
        }
    }


    public class ColumnRightAction extends BriefAction
    {
        public ColumnRightAction()
        {
            super(TXT_RIGHT, "");
        }

        protected void onAction(EditorPane ed)
        {
//            System.err.println("right");
            moveSelection(1, 0, ed);
        }
    }


    public class EscapeAction extends BriefAction
    {
        public EscapeAction()
        {
            super(TXT_ESCAPE, "");
        }

        protected void onAction(EditorPane ed)
        {
//            System.err.println("escape");
            cancelSelection();
        }
    }


    /**
     * Base class for all operations that do something with the selection block.
     */
    public class BlockOperation extends BriefAction
            implements BlockParserListener
    {
        public BlockOperation(String name)
        {
            super(name, "");
        }

        protected void onAction(EditorPane ed)
        {
            parseBlock(ed, this);
        }

        public String lineRead(String line, int required, int actual)
        {
            return null;
        }
    }


    public class ColumnCopyAction extends BlockOperation
    {
        private StringBuffer sb;

        public ColumnCopyAction()
        {
            super("copy");
        }

        public ColumnCopyAction(String s)
        {
            super(s);
        }

        protected void onAction(EditorPane ed)
        {
            sb = new StringBuffer();
            super.onAction(ed);
            cancelSelection();
            Clipboard cli = Toolkit.getDefaultToolkit().getSystemClipboard();
            clipText = sb.toString();
            ColumnStringSelection ss = new ColumnStringSelection(clipText);
            cli.setContents(ss, ss);
            clipOwner = true;
        }

        public String lineRead(String line, int required, int actual)
        {
            sb.append(line);
            for (; actual < required; actual++)
            {
                sb.append(' ');
            }
            sb.append('\n');
            return null;
        }
    }


    public class ColumnCutAction extends ColumnCopyAction
    {
        public ColumnCutAction()
        {
            super("cut");
        }

        public String lineRead(String line, int required, int actual)
        {
            super.lineRead(line, required, actual);
            return "";
        }

    }


    public class ColumnDeleteAction extends BlockOperation
    {
        public ColumnDeleteAction()
        {
            super("delete");
        }

        protected void onAction(EditorPane ed)
        {
            super.onAction(ed);
            cancelSelection();
        }

        public String lineRead(String line, int required, int actual)
        {
            return "";
        }
    }


    public class ColumnUpperCaseAction extends BlockOperation
    {
        public ColumnUpperCaseAction()
        {
            super("uppercase");
        }

        public String lineRead(String line, int required, int actual)
        {
            return line.toUpperCase();
        }
    }


    public class ColumnLowerCaseAction extends BlockOperation
    {
        public ColumnLowerCaseAction()
        {
            super("lowercase");
        }

        public String lineRead(String line, int required, int actual)
        {
            return line.toLowerCase();
        }
    }


    public class ColumnToggleCaseAction extends BlockOperation
    {
        public ColumnToggleCaseAction()
        {
            super("togglecase");
        }

        public String lineRead(String line, int required, int actual)
        {
            char[] chars = line.toCharArray();
            for (int i = 0; i < chars.length; i++)
            {
                char c = chars[i];
                chars[i] = Character.isLowerCase(c)
                            ? Character.toUpperCase(c)
                            : Character.toLowerCase(c);
            }
            return new String(chars);
        }
    }

/*
    public class BufferListAction extends BriefAction
    {
        public BufferListAction()
        {
            super("buffer-list", "Show list of open buffers.");
        }

        public void actionPerformed( ActionEvent ev )
        {
            Browser br = Browser.getActiveBrowser();
            Node[] nodes = br.getOpenNodes();
            if ( nodes.length==0 )
                return;
            Node current = br.getActiveNode();
            String[] names = new String[ nodes.length ];
            String select = null;
            HashMap h = new HashMap();
            for ( int i=0; i<nodes.length; i++ )
            {
                names[i] = nodes[i].getLongDisplayName();
                h.put( names[i], nodes[i] );
                if ( current == nodes[i] )
                    select = names[i];
            }
            Arrays.sort( names );
            String n = (String)ListChoice.getChoice( "Select buffer to edit",
                                                    names, select );
            if ( n!=null )
            {
                try
                {
                    br.setActiveNode( (Node)h.get( n ), true );
                }
                catch ( Exception ex )
                {
                    System.err.println( "Cannot set node \"" + n +
                    "\" as the active one" );
                }
            }
        }
    }
*/

    /**
     * To copy and paste columnar text inside JB clipboard is not used. The text is stored locally
     * inside the class. Because of thet we need a method for tracing when our stuff is removed from
     * the clipboard.
     */
    private class ColumnStringSelection extends StringSelection
    {
        ColumnStringSelection(String s)
        {
            super(s);
        }

        public void lostOwnership(Clipboard c, Transferable t)
        {
//            System.err.println("lostOwnership");
            clipOwner = false;
        }
    }


    class BlockHighlightPainter extends LayeredHighlighter.LayerPainter
    {
        private Color color = Color.green;

        BlockHighlightPainter(JTextComponent comp)
        {
            MasterStyleContext styleCtx = EditorManager.getStyleContext();
            Style style = styleCtx.getStyle(TXT_COLOR);
            //put defaults for style, no listener required here
            color = StyleConstants.getBackground(style);
        }

        public Shape paintLayer(Graphics g,
                                int p0,
                                int p1,
                                Shape s,
                                JTextComponent editor,
                                View v)
        {
            try
            {
                r.x = Math.min(startColumn, column) * cx + leftMargin;
                r.width = (Math.abs(startColumn - column)+1) * cx;
                r.y = Math.min(startRow, row) * cy + topMargin;
                r.height = (Math.abs(startRow - row)+1) * cy;

                g.setColor(color);

                Rectangle r1 = editor.modelToView(p0).union(editor.modelToView(p1));
                r.y = Math.max(r.y, r1.y);
                r.height = Math.min(r.height, r1.height);

                g.fillRect(r.x, r.y, r.width, r.height);
                return r;
            }
            catch (BadLocationException ex)
            {
                ex.printStackTrace();
            }
            return null;
        }


        public void paint(Graphics g, int offs0, int offs1, Shape s, JTextComponent c)
        {
                //System.err.println("paint() called"); //never happens
        }
    }


}

/*
private class ListChoice extends JDialog
{
    private JList list = new JList();
    private Object choice;
    private Container cp;

    private ListChoice( Frame parent, String title,
                        Object[] elements, Object select )
    {
        super( parent, title, true );
        cp = getContentPane();
        cp.setLayout( new BorderLayout() );
        list.setBorder( new EtchedBorder() );
        list.addKeyListener( new KeyAdapter()
            {
                public void keyPressed( KeyEvent ev )
                {
                    if ( ev.getKeyCode()==KeyEvent.VK_ESCAPE )
                    {
                        dispose();
                        return;
                    }

                    if ( ev.getKeyCode()!=KeyEvent.VK_ENTER
                            && ev.getKeyCode()!=KeyEvent.VK_SPACE )
                        return;
                    choice = list.getSelectedValue();
                    if ( choice==null )
                        return;
                    dispose();
                }
            } );
        DefaultListModel dlm = new DefaultListModel();
        for ( int i=0; i<elements.length; dlm.addElement(
                                        elements[i++].toString() ) );
        list.setModel( dlm );
        if ( select != null )
            list.setSelectedValue( select, true );

        list.setBorder( new EmptyBorder(5,5,5,5) );
        //Font f = list.getFont();
        //list.setFont( new Font( f.getFontName(), Font.BOLD, f.getSize() ) );
        cp.add( list, BorderLayout.CENTER );
        pack();
        center();
    }

    private void center()
    {
        Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
        //!!!
        screenSize.width -= 30;
        screenSize.height -= 30;
        Dimension frameSize = getSize();
        frameSize.height = Math.max( screenSize.height/7, frameSize.height );
        frameSize.width = Math.max( screenSize.width/7, frameSize.width );
        if ( frameSize.height > screenSize.height
            || frameSize.width > (screenSize.width*2/3) )
        {
            frameSize.height = Math.min( screenSize.height, frameSize.height );
            frameSize.width = Math.min( screenSize.width*2/3, frameSize.width );
            cp.remove( list );
            JScrollPane scroll = new JScrollPane( list );
            cp.add( scroll, BorderLayout.CENTER );
        }
        setBounds( (screenSize.width - frameSize.width)/2,
                        (screenSize.height - frameSize.height)/2,
                         frameSize.width+30, frameSize.height+30 );
    }

    public static Object getChoice( String title,
                                    Object[] elements, Object select )
    {
        ListChoice dlg = new ListChoice( Browser.getActiveBrowser(),
                                        title, elements, select );
        dlg.show();
        return dlg.choice;
    }
}


class NoCaseComparator implements Comparator
{
    public int compare(Object o1, Object o2)
    {
        String s1 = o1.toString().toLowerCase();
        String s2 = o2.toString().toLowerCase();
        return s1.compareTo( s2 );
    }

    public boolean equals(Object obj)
    {
        return this==obj;
    }
}

*/
